from typing import Any, Callable, Optional, Sequence

import haiku as hk
import jax
import jax.numpy as jnp

from ._nn import sgu_factory
from ._types import XModuleFactory
from ._utils import group_by_prefix_and_trim


class MLPMixerXPatchFeedForward(hk.Module):
    """Patch (token) mixing feedforward layer as described in *MLP-Mixer: An all-MLP Architecture for Vision*¹.

    Note that this module does not implement normalization nor a skip connection. This module can be combined with the
    `XSublayer` module to add these functionalities.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension. Not used directly, rather it's included as an argument to establish
            a consistent interface with other modules.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        dim_hidden (int, optional): Hidden dimension size. Defaults to 4 x num_patches.
        activation (Callable[[jnp.ndarray], jnp.ndarray], optional): Activation function. Defaults to the GELU function.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Currently unused.

    References:
        1. MLP-Mixer: An all-MLP Architecture for Vision (https://arxiv.org/abs/2105.01601).
    """

    def __init__(
        self,
        num_patches: int,
        dim: int,
        depth: int,
        dim_hidden: Optional[int] = None,
        activation: Callable[[jnp.ndarray], jnp.ndarray] = jax.nn.gelu,
        name: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(name=name)

        if kwargs:
            raise KeyError(f"unknown keyword arguments: {list(kwargs.keys())}")

        self.num_patches = num_patches
        self.dim = dim
        self.depth = depth
        self.dim_hidden = num_patches * 4 if dim_hidden is None else dim_hidden
        self.activation = activation

    def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
        # NOTE: Equivalent to a transpose and pair of linear layers.
        x = hk.Conv1D(self.dim_hidden, 1, data_format="NCW", name="conv_1")(inputs)
        x = self.activation(x)
        return hk.Conv1D(self.num_patches, 1, data_format="NCW", name="conv_2")(x)


class ResMLPXPatchFeedForward(hk.Module):
    """Patch (token) mixing feedforward layer as described in ResMLP¹.

    Note that this module does not implement normalization nor a skip connection. This module can be combined with the
    `XSublayer` module to add these functionalities.

    Args
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension. Not used directly, rather it's included as an argument to establish
            a consistent interface with other modules.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        **kwargs: Currently unused.

    References:
        1. ResMLP: Feedforward networks for image classification with data-efficient training
            (https://arxiv.org/abs/2105.03404).
    """

    def __init__(self, num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any):
        super().__init__(name=name)

        if kwargs:
            raise KeyError(f"unknown keyword arguments: {list(kwargs.keys())}")

        self.num_patches = num_patches
        self.dim = dim
        self.depth = depth

    def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
        # NOTE: Equivalent to a transpose and linear layer.
        return hk.Conv1D(self.num_patches, 1, data_format="NCW", name="conv")(inputs)


class gMLPFeedForward(hk.Module):
    """gMLP feedforward layer as described in *Pay Attention to MLPs*¹.

    Note that this module does not implement normalization nor a skip connection. This module can be combined with the
    `XSublayer` module to add these functionalities.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        dim_hidden (int, optional): Hidden dimension size of the first projection. Defaults to 4 x dim.
        sgu (XModuleFactory, optional): Spatial gating unit factory function. Defaults to the standard SpatialGatingUnit
            module via `sgu_factory`.
        activation (Callable[[jnp.ndarray], jnp.ndarray], optional): Activation function. Defaults to the GELU function.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: All arguments starting with "sgu_" are passed to the spatial gating unit factory function.

    References:
        1. Pay Attention to MLPs (https://arxiv.org/abs/2105.08050).
    """

    def __init__(
        self,
        num_patches: int,
        dim: int,
        depth: int,
        dim_hidden: Optional[int] = None,
        sgu: Optional[XModuleFactory] = None,
        activation: Callable[[jnp.ndarray], jnp.ndarray] = jax.nn.gelu,
        name: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(name=name)

        sgu_kwargs, kwargs = group_by_prefix_and_trim("sgu_", kwargs)
        if kwargs:
            raise KeyError(f"unknown keyword arguments: {list(kwargs.keys())}")

        if sgu is None:
            sgu = sgu_factory

        self.num_patches = num_patches
        self.dim = dim
        self.depth = depth
        self.dim_hidden = dim * 4 if dim_hidden is None else dim_hidden
        self.sgu = sgu
        self.activation = activation
        self.sgu_kwargs = sgu_kwargs

    def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
        x = hk.Linear(self.dim_hidden, name="proj_hidden")(inputs)
        x = self.activation(x)
        x = self.sgu(self.num_patches, self.dim, self.depth, **self.sgu_kwargs, name="sgu")(x)
        x = hk.Linear(self.dim, name="proj_dim")(x)
        return x


class FFTFeedForward(hk.Module):
    """N-D FFT mixing feedforward layer.

    Mixes patches (tokens) using an N-D FFT. By default this is done in the patch/token/sequence dimension using a 1D FFT.
    Returns only the real component of the FFT.

    Note that this module does not implement normalization nor a skip connection. This module can be combined with the
    `XSublayer` module to add these functionalities.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        axes (Sequence[int]): The axes to perform the FFT on. Defaults to (-2,).
        norm (str, optional): The normalization to apply to the FFT. Defaults to None.
        name (str, optional): The name of the module. Defaults to None.
    """

    def __init__(
        self,
        num_patches: int,
        dim: int,
        depth: int,
        axes: Sequence[int] = (-2,),
        norm: Optional[str] = None,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)

        self.num_patches = num_patches
        self.dim = dim
        self.depth = depth
        self.axes = axes
        self.norm = norm

    def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
        return jnp.fft.fftn(inputs, axes=self.axes, norm=self.norm).real


class FFTLinearFeedForward(hk.Module):
    """1-D FFT mixing combined with a linear layer.

    Mixes patches (tokens) using an 1-D FFT across the patch/token/sequence dimension and then feeds the results to a
    linear layer. Operates only on the real component of the FFT.

    Note that this module does not implement normalization nor a skip connection. This module can be combined with the
    `XSublayer` module to add these functionalities.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        axis (int): The axes to perform the FFT on. Defaults to (-2,).
        norm (str, optional): The normalization to apply to the FFT. Defaults to None.
        name (str, optional): The name of the module. Defaults to None.
    """

    def __init__(
        self,
        num_patches: int,
        dim: int,
        depth: int,
        axis: int = -2,
        norm: Optional[str] = None,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)

        self.num_patches = num_patches
        self.dim = dim
        self.depth = depth
        self.axis = axis
        self.norm = norm

    def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
        x = jnp.fft.fft(inputs, axis=self.axis, norm=self.norm).real
        return hk.Linear(self.dim, name="linear")(x)


class XChannelFeedForward(hk.Module):
    """Common channel mixing feedforward layer.

    Note that this module does not implement normalization nor a skip connection. This module can be combined with the
    `XSublayer` module to add these functionalities.

    Args:
        num_patches (int): Number of patches in the input. Not used directly, rather it's included as an argument to
            establish a consistent interface with other modules.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        dim_hidden (int, optional): Hidden dimension size. Defaults to 4 x dim.
        activation (Callable[[jnp.ndarray], jnp.ndarray], optional): Activation function. Defaults to the GELU function.
        shift (Callable[[jnp.ndarray], jnp.ndarray], optional): Token shifting function (e.g., spatial shift in S^2-MLP).
            Defaults to `None`.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Currently unused.
    """

    def __init__(
        self,
        num_patches: int,
        dim: int,
        depth: int,
        dim_hidden: Optional[int] = None,
        activation: Callable[[jnp.ndarray], jnp.ndarray] = jax.nn.gelu,
        shift: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
        name: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(name=name)

        if kwargs:
            raise KeyError(f"unknown keyword arguments: {list(kwargs.keys())}")

        self.num_patches = num_patches
        self.dim = dim
        self.depth = depth
        self.dim_hidden = dim * 4 if dim_hidden is None else dim_hidden
        self.activation = activation
        self.shift = shift

    def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
        x = hk.Linear(self.dim_hidden, name="linear_1")(inputs)
        x = self.activation(x)
        if self.shift is not None:
            x = self.shift(x)
        x = hk.Linear(self.dim, name="linear_2")(x)
        return x


def gmlp_feedforward_factory(
    num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any
) -> hk.Module:
    """gMLP feedforward module factory function.

    Satifies the `XModuleFactory` interface.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Additional `gMLPFeedForward` arguments.

    Returns:
        hk.Module: `gMLPFeedForward` module.
    """

    return gMLPFeedForward(num_patches, dim, depth, **kwargs, name=name)


def mlpmixer_xpatch_feedforward_factory(
    num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any
) -> hk.Module:
    """MLP-Mixer cross-patch feedforward module factory function.

    Satifies the `XModuleFactory` interface.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Additional `MLPMixerXPatchFeedForward` arguments.

    Returns:
        hk.Module: `MLPMixerXPatchFeedForward` module.
    """

    return MLPMixerXPatchFeedForward(num_patches, dim, depth, **kwargs, name=name)


def resmlp_xpatch_feedforward_factory(
    num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any
) -> hk.Module:
    """ResMLP cross-patch feedforward module factory function.

    Satifies the `XModuleFactory` interface.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Additional `ResMLPXPatchFeedForward` arguments.

    Returns:
        hk.Module: `ResMLPXPatchFeedForward` module.
    """
    return ResMLPXPatchFeedForward(num_patches, dim, depth, **kwargs, name=name)


def fft_xpatch_feedforward_factory(
    num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any
) -> hk.Module:
    """FFT cross-patch feedforward module factory function.

    Satifies the `XModuleFactory` interface.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Additional `FFTFeedForward` arguments.

    Returns:
        hk.Module: `FFTFeedForward` module.
    """
    return FFTFeedForward(num_patches, dim, depth, **kwargs, name=name)


def fftlinear_xpatch_feedforward_factory(
    num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any
) -> hk.Module:
    """FFT-Linear cross-patch feedforward module factory function.

    Satifies the `XModuleFactory` interface.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Additional `FFTLinearFeedForward` arguments.

    Returns:
        hk.Module: `FFTLinearFeedForward` module.
    """
    return FFTLinearFeedForward(num_patches, dim, depth, **kwargs, name=name)


def xchannel_feedforward_factory(
    num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any
) -> hk.Module:
    """Standard cross-channel feedforward module factory function.

    Satifies the `XModuleFactory` interface.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Additional `XChannelFeedForward` arguments.

    Returns:
        hk.Module: `XChannelFeedForward` module.
    """
    return XChannelFeedForward(num_patches, dim, depth, **kwargs, name=name)


__all__ = [
    "MLPMixerXPatchFeedForward",
    "FFTFeedForward",
    "FFTLinearFeedForward",
    "ResMLPXPatchFeedForward",
    "XChannelFeedForward",
    "fft_xpatch_feedforward_factory",
    "fftlinear_xpatch_feedforward_factory",
    "gMLPFeedForward",
    "gmlp_feedforward_factory",
    "mlpmixer_xpatch_feedforward_factory",
    "resmlp_xpatch_feedforward_factory",
    "xchannel_feedforward_factory",
]



from typing import Any, Optional, Protocol

import haiku as hk


class XModuleFactory(Protocol):
    """Defines a common factory function interface for all X-MLP modules."""

    def __call__(self, num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any) -> hk.Module:
        ...


__all__ = ["XModuleFactory"]



from typing import Any, Optional, Sequence, Union

import haiku as hk
import jax.numpy as jnp
from einops import reduce

from ._ff import (
    fft_xpatch_feedforward_factory,
    fftlinear_xpatch_feedforward_factory,
    gmlp_feedforward_factory,
    mlpmixer_xpatch_feedforward_factory,
    resmlp_xpatch_feedforward_factory,
    xchannel_feedforward_factory,
)
from ._nn import Affine, DropPath, LayerScale, layernorm_factory
from ._types import XModuleFactory
from ._utils import group_by_prefix_and_trim


class XSublayer(hk.Module):
    """Flexible sublayer wrapper module providing skip connections and pre/post-normalization to arbitrary layers.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        ff (XModuleFactory): Feedforward layer factory function.
        prenorm (XModuleFactory, optional): Pre-normalization layer factory function. Defaults to `None`.
        postnorm (XModuleFactory, optional): Post-normalization layer factory function. Defaults to `None`.
        residual (bool): Whether to add a residual/skip connection. Defaults to `True`.
        p_drop_path (float): Probability of the core computation being dropped. Only applicable if `residual` is `True`.
            Defaults to 0.0.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: All arguments starting with "ff_" are passed to the feedforward layer factory function.
            All arguments starting with "prenorm_" are passed to the pre-normalization layer factory function.
            All arguments starting with "postnorm_" are passed to the post-normalization layer factory function.
    """

    def __init__(
        self,
        num_patches: int,
        dim: int,
        depth: int,
        ff: XModuleFactory,
        prenorm: Optional[XModuleFactory] = None,
        postnorm: Optional[XModuleFactory] = None,
        residual: bool = True,
        p_drop_path: float = 0.0,
        drop_path_mode: str = "batch",
        name: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(name=name)

        ff_kwargs, kwargs = group_by_prefix_and_trim("ff_", kwargs)
        prenorm_kwargs, kwargs = group_by_prefix_and_trim("prenorm_", kwargs)
        postnorm_kwargs, kwargs = group_by_prefix_and_trim("postnorm_", kwargs)
        if kwargs:
            raise KeyError(f"unknown keyword arguments: {list(kwargs.keys())}")

        self.num_patches = num_patches
        self.dim = dim
        self.depth = depth
        self.ff = ff
        self.prenorm = prenorm
        self.postnorm = postnorm
        self.residual = residual
        self.p_drop_path = p_drop_path
        self.drop_path_mode = drop_path_mode
        self.ff_kwargs = ff_kwargs
        self.prenorm_kwargs = prenorm_kwargs
        self.postnorm_kwargs = postnorm_kwargs

    def __call__(self, inputs: jnp.ndarray, *, is_training: bool) -> jnp.ndarray:
        """Propogate inputs through the sublayer.

        Args:
            inputs (jnp.ndarray): Inputs to the sublayer.
            is_training (bool): If `True`, enable training specific features (e.g., dropout). Keyword argument.

        Returns:
            jnp.ndarray: Outputs of the sublayer.
        """
        x = inputs
        if self.prenorm is not None:
            x = self.prenorm(self.num_patches, self.dim, self.depth, **self.prenorm_kwargs)(x)
        x = self.ff(self.num_patches, self.dim, self.depth, **self.ff_kwargs)(x)
        if self.postnorm is not None:
            x = self.postnorm(self.num_patches, self.dim, self.depth, **self.postnorm_kwargs)(x)
        if self.residual:
            x = DropPath(self.p_drop_path, mode=self.drop_path_mode)(x, is_training=is_training) + inputs
        return x


class XBlock(hk.Module):
    """Generic MLP block.

    One or more `XSublayer` modules are stacked together to form a block. Optionally, a skip connection can be added.
    Arbitrary arguments can be passed to `XSublayer` modules two different ways:

    1. As keyword arguments prefixed with "sublayers_". These arguments are passed to all the sublayers.
    2. As keyword arguments prefixed with "sublayer{i}_" where 1 <= i <= len(sublayers). These arguments are passed to
        to the i-th sublayer.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of this block in the network. Note that depth starts from 1.
        sublayers (Sequence[XSublayerFactory]): Sublayer factory functions. Created sublayers will be stacked in the
            order of their respective factory function in the sequence.
        residual (bool): Whether to add a residual/skip connection. Defaults to `False`.
        p_drop_path (float): Probability of the core computation being dropped. Passed directly
            to sublayers. This will also be applied at the block level if residual is `True`. Defaults to 0.0.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: All arguments starting with "sublayers_" are passed to all sublayers. All arguments starting with
            "sublayer{i}_" are passed to the i-th sublayer.
    """

    def __init__(
        self,
        num_patches: int,
        dim: int,
        depth: int,
        sublayers: Sequence[XModuleFactory],
        residual: bool = False,
        p_drop_path: float = 0.0,
        drop_path_mode: str = "batch",
        name: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(name=name)

        sublayer_common_kwargs, kwargs = group_by_prefix_and_trim("sublayers_", kwargs)
        sublayers_kwargs = []
        for i in range(len(sublayers)):
            sublayer_kwargs, kwargs = group_by_prefix_and_trim(f"sublayer{i + 1}_", kwargs)
            sublayers_kwargs.append(sublayer_kwargs)
        if kwargs:
            raise KeyError(f"unknown keyword arguments: {list(kwargs.keys())}")

        self.num_patches = num_patches
        self.dim = dim
        self.depth = depth
        self.sublayers = tuple(sublayers)
        self.residual = residual
        self.p_drop_path = p_drop_path
        self.drop_path_mode = drop_path_mode
        self.sublayer_common_kwargs = sublayer_common_kwargs
        self.sublayers_kwargs = sublayers_kwargs

    def __call__(self, inputs: jnp.ndarray, *, is_training: bool) -> jnp.ndarray:
        """Propogate inputs through the block.

        Args:
            inputs (jnp.ndarray): Inputs to the block.
            is_training (bool): If `True`, enable training specific features (e.g., dropout). Keyword argument.

        Returns:
            jnp.ndarray: Outputs of the block.
        """
        x = inputs
        for i, sublayer in enumerate(self.sublayers):
            sublayer_kwargs = self.sublayer_common_kwargs.copy()
            sublayer_kwargs.update(self.sublayers_kwargs[i])
            x = sublayer(
                self.num_patches,
                self.dim,
                self.depth,
                p_drop_path=self.p_drop_path,
                drop_path_mode=self.drop_path_mode,
                **sublayer_kwargs,
            )(x, is_training=is_training)
        if self.residual:
            x = DropPath(self.p_drop_path, mode=self.drop_path_mode)(x, is_training=is_training) + inputs
        return x


class XMLP(hk.Module):
    """Generic MLP network.

    N `XBlock` modules are stacked together to form a network (where N is set to `depth`). Importantly, this network
    assumes the input has been formatted appropriately (e.g., a sequence of patches). Before data is processed by the
    stack of `XBlock` modules, it is first projected to the specified dimension `dim` via a linear layer.

    This network can optionally be configured with Stochastic Depth¹, a form of regularization. If enabled, the depth
    of the network will be dynamically adjusted during training, with sections of the network being randomly dropped.
    The likelihood of dropping a layer can either fixed, or dependent on the depth of the network.

    Optionally, the network can be configured to have a classification layer at the end by setting `num_classes` to a
    non-zero value. In this case, the resulting sequence from stack of `XBlock` modules will be averaged over the
    sequence dimension before being fed to the classification layer.

    Arbitrary arguments can be passed to `XBlock` modules by prepending the argument name with "block_". Further, to
    ensure arguments are passed to child modules of each `XBlock` module, each argument name should additionally be
    prefixed with that child module's identifier, starting with "block" and working down the hierarchy. For example,
    to pass an argument to the feedforward module of the first sublayer of each block, the argument name should be
    "block_sublayer1_ff_<argument name>".

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension. Inputs fed to this network are projected to this dimension.
        depth (int): The number of blocks in the network.
        block (XBlockFactory): Block factory function.
        normalization (XModuleFactory, optional): Normalization module factory function. Occurs after the stack of
            `XBlock` modules. Useful for pre-normalization architectures. Defaults to None.
        stochastic_depth (Union[bool, float], optional): Whether to use stochastic depth. If `True`, the surivival rate
            of each block follows the linear decay function 1 - 0.5 * (i / depth) for 1 <= i <= depth. If `False`, the
            survival rate is 1.0. If a float, the survival rate is set to this value. Defaults to False.
        num_classes (int, optional): Number of classes in the classification layer. Defaults to None.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: All arguments starting with "block_" are passed to all blocks.

    References:
        1. Deep Networks with Stochastic Depth (https://arxiv.org/abs/1603.09382).
    """

    def __init__(
        self,
        num_patches: int,
        dim: int,
        depth: int,
        block: XModuleFactory,
        normalization: Optional[XModuleFactory] = None,
        stochastic_depth: Union[float, bool] = False,
        num_classes: Optional[int] = None,
        name: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(name=name)

        block_kwargs, kwargs = group_by_prefix_and_trim("block_", kwargs)
        if kwargs:
            raise KeyError(f"unknown keyword arguments: {list(kwargs.keys())}")

        if isinstance(stochastic_depth, bool) and stochastic_depth:
            # This ensures that the first block can be dropped as well.
            p_drop_path = jnp.linspace(0.0, 0.5, num=depth)
        elif isinstance(stochastic_depth, float):
            p_drop_path = jnp.full(depth, stochastic_depth)
        else:
            p_drop_path = jnp.zeros(depth)

        self.num_patches = num_patches
        self.dim = dim
        self.depth = depth
        self.block = block
        self.normalization = normalization
        self.p_drop_path = p_drop_path
        self.num_classes = num_classes
        self.block_kwargs = block_kwargs

    def __call__(self, inputs: jnp.ndarray, *, is_training: bool) -> jnp.ndarray:
        """Propogate inputs through the network.

        Args:
            inputs (jnp.ndarray): Inputs to the network.
            is_training (bool): If `True`, enable training specific features (e.g., dropout). Keyword argument.

        Returns:
            jnp.ndarray: Outputs of the network.
        """
        x = hk.Linear(self.dim, name="proj_in")(inputs)
        for i in range(self.depth):
            x = self.block(
                self.num_patches,
                self.dim,
                i + 1,
                p_drop_path=self.p_drop_path[i],
                **self.block_kwargs,
            )(x, is_training=is_training)
        if self.normalization is not None:
            x = self.normalization(self.num_patches, self.dim, self.depth + 1)(x)
        if self.num_classes is not None:
            x = reduce(x, "... n c -> ... c", "mean")
            x = hk.Linear(self.num_classes, name="proj_out")(x)
        return x


def gmlp_block_factory(num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any) -> hk.Module:
    """gMLP block module factory function.

    Builds a `XBlock` module with the gMLP block structure as defined in *Pay Attention to MLPs*¹. Specifically, this
    consists of a single `XSublayer` with a `gMLPFeedForward` module and layer normalization (pre-normalization).

    Satifies the `XModuleFactory` interface.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Additional block and child module arguments.

    Returns:
        hk.Module: `XBlock` module.

    References:
        1. Pay Attention to MLPs (https://arxiv.org/abs/2105.08050).
    """
    return XBlock(
        num_patches,
        dim,
        depth,
        [
            lambda num_patches, dim, depth, **kwargs: XSublayer(
                num_patches, dim, depth, ff=gmlp_feedforward_factory, prenorm=layernorm_factory, **kwargs
            )
        ],
        name=name,
        **kwargs,
    )


def mlpmixer_block_factory(
    num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any
) -> hk.Module:
    """MLP-Mixer block module factory function.

    Builds a `XBlock` module with the MLP-Mixer block structure as defined in *MLP-Mixer: An all-MLP Architecture for
    Vision*¹. Specifically, this consists of two `XSublayer`s: 1) a `MLPMixerXPatchFeedForward` module and 2) a
    `XChannelFeedForward` module. Both make use of layer normalization (pre-normalization).

    Satifies the `XModuleFactory` interface.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Additional block and child module arguments.

    Returns:
        hk.Module: `XBlock` module.

    References:
        1. MLP-Mixer: An all-MLP Architecture for Vision (https://arxiv.org/abs/2105.01601).
    """
    return XBlock(
        num_patches,
        dim,
        depth,
        [
            # Cross patch sublayer
            lambda num_patches, dim, depth, **kwargs: XSublayer(
                num_patches, dim, depth, ff=mlpmixer_xpatch_feedforward_factory, prenorm=layernorm_factory, **kwargs
            ),
            # Cross channel sublayer
            lambda num_patches, dim, depth, **kwargs: XSublayer(
                num_patches, dim, depth, ff=xchannel_feedforward_factory, prenorm=layernorm_factory, **kwargs
            ),
        ],
        name=name,
        **kwargs,
    )


def resmlp_block_factory(
    num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any
) -> hk.Module:
    """ResMLP block module factory function.

    Builds a `XBlock` module with the ResMLP block structure as defined in *ResMLP: Feedforward networks for image
    classification with data-efficient training*¹. Specifically, this consists of two `XSublayer`s: 1) a
    `ResMLPXPatchFeedForward` module and 2) a `XChannelFeedForward` module. Both make use of `Affine` pre-normalization
    and `LayerScale` post-normalization.

    Satifies the `XModuleFactory` interface.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Additional block and child module arguments.

    Returns:
        hk.Module: `XBlock` module.

    References:
        1. ResMLP: Feedforward networks for image classification with data-efficient training
            (https://arxiv.org/abs/2105.03404).
    """
    return XBlock(
        num_patches,
        dim,
        depth,
        [
            lambda num_patches, dim, depth, **kwargs: XSublayer(
                num_patches,
                dim,
                depth,
                ff=resmlp_xpatch_feedforward_factory,
                prenorm=lambda num_patches, dim, depth, **kwargs: Affine(dim, **kwargs),
                postnorm=lambda num_patches, dim, depth, **kwargs: LayerScale(dim, depth, **kwargs),
                **kwargs,
            ),
            lambda num_patches, dim, depth, **kwargs: XSublayer(
                num_patches,
                dim,
                depth,
                ff=xchannel_feedforward_factory,
                prenorm=lambda num_patches, dim, depth, **kwargs: Affine(dim, **kwargs),
                postnorm=lambda num_patches, dim, depth, **kwargs: LayerScale(dim, depth, **kwargs),
                **kwargs,
            ),
        ],
        name=name,
        **kwargs,
    )


def s2mlp_block_factory(num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any) -> hk.Module:
    """S^2-MLP block module factory function.

    Builds a `XBlock` module with the S^2-MLP block structure as defined in *S^2-MLP: Spatial-Shift MLP Architecture
    for Vision*¹. Specifically, this consists of two `XSublayer`s: 1) a `XChannelFeedForward` module with a shift
    function and 2) a second `XChannelFeedForward` module. Both make use of layer normalization (pre-normalization).

    Note: currently, the spatial shift function must be passed as the keyword argument "sublayer1_ff_shift". This is
    due to the fact that `create_shift2d_op` requires post-patching height and width dimensions, which the `XMLP`
    module has no knowledge of.

    Satifies the `XModuleFactory` interface.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Additional block and child module arguments.

    Returns:
        hk.Module: `XBlock` module.

    References:
        1. S2-MLP: Spatial-Shift MLP Architecture for Vision (https://arxiv.org/abs/2106.07477).
    """

    if "sublayer1_ff_shift" not in kwargs:
        raise ValueError("s2mlp_block_factory requires sublayer1_ff_shift to be specified")
    if "sublayer1_ff_dim_hidden" not in kwargs and "sublayers_ff_dim_hidden" not in kwargs:
        kwargs["sublayer1_ff_dim_hidden"] = dim

    return XBlock(
        num_patches,
        dim,
        depth,
        [
            # Cross patch sublayer
            lambda num_patches, dim, depth, **kwargs: XSublayer(
                num_patches, dim, depth, ff=xchannel_feedforward_factory, prenorm=layernorm_factory, **kwargs
            ),
            # Cross channel sublayer
            lambda num_patches, dim, depth, **kwargs: XSublayer(
                num_patches, dim, depth, ff=xchannel_feedforward_factory, prenorm=layernorm_factory, **kwargs
            ),
        ],
        name=name,
        **kwargs,
    )


def fft_block_factory(num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any) -> hk.Module:
    """FFT block module factory function.

    Builds a `XBlock` module with an FNet-like block structure as defined in *FNet: Mixing Tokens with Fourier Transforms*¹.
    Specifically, this consists of two `XSublayer`s: 1) an `FFTFeedForward` module and 2) a `XChannelFeedForward` module.
    Both make use of layer normalization (pre-normalization).

    In contrast to the FNet paper, the FFT feedforward layer can be configured to an arbitrary N-D FFT. By default it is
    configured to be a 1D FFT across the patch (token) dimension.

    Satifies the `XModuleFactory` interface.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Additional block and child module arguments.

    Returns:
        hk.Module: `XBlock` module.

    References:
        1. FNet: Mixing Tokens with Fourier Transforms (https://arxiv.org/abs/2105.03824).
    """
    return XBlock(
        num_patches,
        dim,
        depth,
        [
            # Cross patch sublayer
            lambda num_patches, dim, depth, name=None, **kwargs: XSublayer(
                num_patches,
                dim,
                depth,
                ff=fft_xpatch_feedforward_factory,
                prenorm=layernorm_factory,
                name=name,
                **kwargs,
            ),
            # Cross channel sublayer
            lambda num_patches, dim, depth, name=None, **kwargs: XSublayer(
                num_patches, dim, depth, ff=xchannel_feedforward_factory, prenorm=layernorm_factory, name=name, **kwargs
            ),
        ],
        name=name,
        **kwargs,
    )


def fftlinear_block_factory(
    num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any
) -> hk.Module:
    """FFT-Linear block module factory function.

    Builds a `XBlock` module with an FNet-like block structure as defined in *FNet: Mixing Tokens with Fourier Transforms*¹.
    Specifically, this consists of two `XSublayer`s: 1) an `FFTLinearFeedForward` module and 2) a `Linear` module. Both
    make use of layer normalization (pre-normalization).

    In contrast to the FNet paper, the feedforward layer consists of a 1-D FFT across the patch (token) dimension and a
    linear layer (which consumes the FFT'ed data).

    Satifies the `XModuleFactory` interface.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Additional block and child module arguments.

    Returns:
        hk.Module: `XBlock` module.

    References:
        1. FNet: Mixing Tokens with Fourier Transforms (https://arxiv.org/abs/2105.03824).
    """
    return XBlock(
        num_patches,
        dim,
        depth,
        [
            # Cross patch sublayer
            lambda num_patches, dim, depth, name=None, **kwargs: XSublayer(
                num_patches,
                dim,
                depth,
                ff=fftlinear_xpatch_feedforward_factory,
                prenorm=layernorm_factory,
                name=name,
                **kwargs,
            ),
            # Linear sublayer
            lambda num_patches, dim, depth, name=None, **kwargs: XSublayer(
                num_patches, dim, depth, ff=xchannel_feedforward_factory, prenorm=layernorm_factory, name=name, **kwargs
            ),
        ],
        name=name,
        **kwargs,
    )


__all__ = [
    "XBlock",
    "XMLP",
    "XSublayer",
    "fft_block_factory",
    "fftlinear_block_factory",
    "gmlp_block_factory",
    "mlpmixer_block_factory",
    "resmlp_block_factory",
    "s2mlp_block_factory",
]



from ._ff import *  # noqa: F401, F403
from ._nn import *  # noqa: F401, F403
from ._types import XModuleFactory  # noqa: F401
from ._version import __version__  # noqa: F401
from ._x_mlps import *  # noqa: F401, F403



from functools import partial

# Taken from https://github.com/lucidrains/x-transformers/blob/main/x_transformers/x_transformers.py
# Lightly modified - all credit goes to lucidrains


def _string_begins_with(prefix, str):
    return str.startswith(prefix)


def pick_and_pop(keys, d):
    values = list(map(lambda key: d.pop(key, None), keys))
    return values


def group_dict_by_key(cond, d):
    return_val = [dict(), dict()]
    for key in d.keys():
        match = bool(cond(key))
        ind = int(not match)
        return_val[ind][key] = d[key]
    return (*return_val,)


def group_by_key_prefix(prefix, d):
    return group_dict_by_key(partial(_string_begins_with, prefix), d)


def group_by_prefix_and_trim(prefix, d):
    kwargs_with_prefix, kwargs = group_dict_by_key(partial(_string_begins_with, prefix), d)
    kwargs_without_prefix = dict(map(lambda x: (x[0][len(prefix) :], x[1]), tuple(kwargs_with_prefix.items())))
    return kwargs_without_prefix, kwargs


__all__ = ["group_by_key_prefix", "group_by_prefix_and_trim", "group_dict_by_key", "pick_and_pop"]



from typing import Any, Callable, Optional, Sequence

import haiku as hk
import jax
import jax.numpy as jnp

from ._types import XModuleFactory
from ._utils import pick_and_pop


def _calc_layer_scale_eps(depth: int) -> float:
    if depth <= 18:
        init_eps = 0.1
    elif depth > 18 and depth <= 24:
        init_eps = 1e-5
    else:
        init_eps = 1e-6
    return init_eps


def create_multishift1d_op(amount: Sequence[int], bidirectional: bool = True) -> Callable[[jnp.ndarray], jnp.ndarray]:
    """Create a 1D multi-shift operator, inspired by the spatial shift operator introduced in S^2-MLP¹.

    This operation works by calling `create_shift1d_op` with each value in `amount` and then concatenating the results.
    By default the shift operator is bidirectional, resulting in interleaved shifting.

    Expects inputs to be of shape `(num_patches, dim)` or `(batch_size, num_patches, dim)`.

    Args:
        amount: Sequence of shift amounts.
        bidirectional: Whether to use bidirectional shift. Defaults to True.

    Returns:
        The configured shift operator.

    References:
        1. S2-MLP: Spatial-Shift MLP Architecture for Vision (https://arxiv.org/abs/2106.07477).
    """

    def multishift1d(x: jnp.ndarray) -> jnp.ndarray:
        xs = jnp.split(x, len(amount), axis=-1)
        shifted_xs = [create_shift1d_op(amount[i], bidirectional=bidirectional)(x) for i, x in enumerate(xs)]
        x = jnp.concatenate(shifted_xs, axis=-1)
        return x

    return multishift1d


def create_shift1d_op(amount: int = 1, bidirectional: bool = True) -> Callable[[jnp.ndarray], jnp.ndarray]:
    """Create a 1D shift operator, inspired by the spatial shift operator introduced in S^2-MLP¹.

    By default the the shift operator is bidirectional. This results in the input data being split into two parts,
    one part is shifted forward and the other part is shifted backward by `amount`. The two parts are then
    concatenated. It's rare that unidirectional shift is needed; however, it can be useful as a building block for more
    complex shifting operations.

    Expects inputs to be of shape `(..., num_patches, dim)`.

    Args:
        amount: Amount of shift. Defaults to 1.
        bidirectional: Whether to use bidirectional shift. Defaults to True.

    Returns:
        The configured shift operator.

    References:
        1. S2-MLP: Spatial-Shift MLP Architecture for Vision (https://arxiv.org/abs/2106.07477).
    """

    def shift1d(x: jnp.ndarray) -> jnp.ndarray:
        slice_from1 = (slice(None, None, None),) * (x.ndim - 2) + (slice(amount, None, None),)
        slice_end1 = (slice(None, None, None),) * (x.ndim - 2) + (slice(None, -amount, None),)
        if bidirectional:
            x1, x2 = jnp.split(x, 2, axis=-1)
            x1 = x1.at[slice_from1].set(x1[slice_end1])
            x2 = x2.at[slice_end1].set(x2[slice_from1])
            x = jnp.concatenate([x1, x2], axis=-1)
        else:
            x = x.at[slice_from1].set(x[slice_end1])
        return x

    return shift1d


def create_shift2d_op(amount: int = 1) -> Callable[[jnp.ndarray], jnp.ndarray]:
    """Create a 2D shift operator based on spatial shift algorithm introduced in S^2-MLP¹.

    Expects inputs to be of shape `(..., height, width, dim)`.

    Args:
        amount: Amount of shift.

    Returns:
        Callable[[jnp.ndarray], jnp.ndarray]: The configured shift operator.

    References:
        1. S2-MLP: Spatial-Shift MLP Architecture for Vision (https://arxiv.org/abs/2106.07477).
    """

    def shift2d(x: jnp.ndarray) -> jnp.ndarray:
        x1, x2, x3, x4 = jnp.split(x, 4, axis=-1)

        # Handle variable tensor shapes.
        hslice_from1 = (slice(None, None, None),) * (x.ndim - 3) + (slice(amount, None, None),)
        hslice_end1 = (slice(None, None, None),) * (x.ndim - 3) + (slice(None, -amount, None),)
        wslice_from1 = (slice(None, None, None),) * (x.ndim - 2) + (slice(amount, None, None),)
        wslice_end1 = (slice(None, None, None),) * (x.ndim - 2) + (slice(None, -amount, None),)

        x1 = x1.at[hslice_from1].set(x1[hslice_end1])
        x2 = x2.at[hslice_end1].set(x2[hslice_from1])
        x3 = x3.at[wslice_from1].set(x3[wslice_end1])
        x4 = x4.at[wslice_end1].set(x4[wslice_from1])

        x = jnp.concatenate([x1, x2, x3, x4], axis=-1)
        return x

    return shift2d


class DropPath(hk.Module):
    """Randomly drop the input with a given probability.

    This is equivalent to Stochastic Depth when applied to the output of a network path¹. Additionally, individual
    samples can be dropped independently when `mode` is set to `sample` (the default is to drop the entire batch),
    mimicking the implementation from PyTorch Image Models². However, sample mode is not supported with `vmap`.

    Args:
        rate: Probability of dropping an element.
        mode: Whether to operate on individual samples in a batch or the whole batch. Defaults to "batch". Must be set
            "batch" when `vmap` is used.
        name: Name of the module.

    References:
        1. Deep Networks with Stochastic Depth (https://arxiv.org/abs/1603.09382).
        2. PyTorch Image Models (https://github.com/rwightman/pytorch-image-models/blob/e4360e6125bb0bb4279785810c8eb33b40af3ebd/timm/models/layers/drop.py#L137).
    """

    def __init__(self, rate: float, mode: str = "batch", name: Optional[str] = None):
        super().__init__(name=name)

        if mode not in {"batch", "sample"}:
            raise ValueError(f"invalid mode: {mode}")

        self.rate = rate
        self.mode = mode

    def __call__(self, x: jnp.ndarray, *, is_training: bool) -> jnp.ndarray:
        if is_training:
            if self.mode == "batch":
                x = (
                    jax.random.bernoulli(hk.next_rng_key(), 1 - self.rate)
                    * x
                    / jnp.clip(1 - self.rate + 1e-12, a_min=0, a_max=1)
                )
            elif self.mode == "sample":
                shape = (x.shape[0],) + (1,) * (x.ndim - 1)
                x = (
                    jax.random.bernoulli(hk.next_rng_key(), 1 - self.rate, shape=shape)
                    * x
                    / jnp.clip(1 - self.rate + 1e-12, a_min=0, a_max=1)
                )
        return x


class Affine(hk.Module):
    """Affine transform layer as described in ResMLP¹.

    Briefly, this operator rescales and shifts its input by a learned weight and bias of size `dim`.

    Args:
        dim (int): Size of the channel dimension.
        init_scale (float): Initial weight value of alpha.
        name (str, optional): The name of the module. Defaults to None.

    References:
        1. ResMLP: Feedforward networks for image classification with data-efficient training
            (https://arxiv.org/abs/2105.03404).
    """

    def __init__(self, dim: int, init_scale: float = 1.0, name: Optional[str] = None):
        super().__init__(name=name)

        self.dim = dim
        self.init_scale = init_scale

    def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
        shape = (1,) * (inputs.ndim - 1) + (self.dim,)
        w = hk.get_parameter("w", shape=shape, init=hk.initializers.Constant(self.init_scale))
        b = hk.get_parameter("b", shape=shape, init=jnp.zeros)

        return w * inputs + b


class LayerScale(hk.Module):
    """LayerScale layer as described in *Going deeper with Image Transformers*¹.

    Briefly, rescales the input by a learned weight of size `dim`.

    Args:
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. This is used to determine the
            initial weight values. Note that depth starts from 1.
        name (str, optional): The name of the module. Defaults to None.

    References:
        1. Going deeper with Image Transformers (https://arxiv.org/abs/2103.17239).
    """

    def __init__(self, dim: int, depth: int, name: Optional[str] = None):
        super().__init__(name=name)

        self.dim = dim
        self.init_eps = _calc_layer_scale_eps(depth)

    def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
        shape = (1,) * (inputs.ndim - 1) + (self.dim,)
        s = hk.get_parameter("s", shape=shape, init=hk.initializers.Constant(self.init_eps))

        return s * inputs


class SpatialGatingUnit(hk.Module):
    """Spatial Gating Unit as described in *Pay Attention to MLPs*¹.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        norm (XModuleFactory, optional): Normalization layer factory function. Defaults to LayerNorm via
            `layernorm_factory`.
        activation (Callable[[jnp.ndarray], jnp.ndarray], optional): Activation function. Applied to the gate values
            spatial projection. Defaults to the identity function.
        init_eps (float): Initial weight of the spatial projection layer. Scaled by the number of patches. Defaults to
            1e-3.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Currently unused.

    References:
        1. Pay Attention to MLPs (https://arxiv.org/abs/2105.08050).
    """

    def __init__(
        self,
        num_patches: int,
        dim: int,
        depth: int,
        norm: XModuleFactory = None,
        activation: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
        init_eps: float = 1e-3,
        name: Optional[str] = None,
        **kwargs: Any,
    ):
        super().__init__(name=name)

        if kwargs:
            raise KeyError(f"unknown keyword arguments: {list(kwargs.keys())}")

        if norm is None:
            norm = layernorm_factory
        if activation is None:
            activation = lambda x: x  # noqa: E731

        self.num_patches = num_patches
        self.dim = dim
        self.depth = depth
        self.norm = norm
        self.activation = activation
        self.init_eps = init_eps / num_patches

    def __call__(self, inputs: jnp.ndarray) -> jnp.ndarray:
        u, v = jnp.split(inputs, 2, axis=-1)
        v = self.norm(self.num_patches, self.dim, self.depth, name="norm")(v)
        v = hk.Conv1D(
            self.num_patches,
            1,
            w_init=hk.initializers.RandomUniform(-self.init_eps, self.init_eps),
            b_init=jnp.ones,
            data_format="NCW",
            name="conv",
        )(v)
        return u * self.activation(v)


def layernorm_factory(num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any) -> hk.Module:
    """Layer normalization module factory function.

    Standard `hk.LayerNorm` arguments can be passed via the `kwargs` dictionary. If no arguments are passed, the
    following defaults are used:

    1. `axis` is set to `-1` (the last dimension).
    2. `create_scale` is set to `True`.
    3. `create_offset` is set to `True`.

    Satifies the `XModuleFactory` interface.

    Args:
        num_patches (int): Number of patches in the input. Unused.
        dim (int): Size of the channel dimension. Unused.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
            Unused.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: `hk.LayerNorm` arguments.

    Returns:
        hk.Module: `hk.LayerNorm` module.
    """
    axis, create_scale, create_offset = pick_and_pop(["axis", "create_scale", "create_offset"], kwargs)
    if axis is None:
        axis = -1
    if create_scale is None:
        create_scale = True
    if create_offset is None:
        create_offset = True

    return hk.LayerNorm(axis, create_scale, create_offset, **kwargs, name=name)


def sgu_factory(num_patches: int, dim: int, depth: int, name: Optional[str] = None, **kwargs: Any) -> hk.Module:
    """`SpatialGatingUnit` module factory function.

    Satifies the `XModuleFactory` interface.

    Args:
        num_patches (int): Number of patches in the input.
        dim (int): Size of the channel dimension.
        depth (int): The depth of the block which contains this layer in the network. Note that depth starts from 1.
        name (str, optional): The name of the module. Defaults to None.
        **kwargs: Additional `SpatialGatingUnit` arguments.

    Returns:
        hk.Module: `SpatialGatingUnit` module.
    """

    return SpatialGatingUnit(num_patches, dim, depth, **kwargs, name=name)


__all__ = [
    "Affine",
    "LayerScale",
    "DropPath",
    "SpatialGatingUnit",
    "layernorm_factory",
    "create_multishift1d_op",
    "create_shift1d_op",
    "create_shift2d_op",
    "sgu_factory",
]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['x_mlps']

package_data = \
{'': ['*']}

install_requires = \
['dm-haiku>=0.0.6,<0.0.7', 'einops>=0.4.1,<0.5.0']

setup_kwargs = {
    'name': 'x-mlps',
    'version': '0.5.0',
    'description': 'Configurable MLPs built on JAX and Haiku',
    'long_description': '# X-MLPs\n\nAn MLP model that provides a flexible foundation to implement, mix-and-match, and test various state-of-the-art MLP building blocks and architectures.\nBuilt on JAX and Haiku.\n\n## Installation\n\n```sh\npip install x-mlps\n```\n\n**Note**: X-MLPs will not install JAX for you (see [here](https://github.com/google/jax#installation) for install instructions).\n\n## Getting Started\n\nThe `XMLP` class provides the foundation from which all MLP architectures are built on, and is the primary class you use.\nAdditionally, X-MLPs relies heavily on factory functions to customize and instantiate the building blocks that make up a particular `XMLP` instance.\nFortunately, this library provides several SOTA MLP blocks out-of-the-box as factory functions.\nFor example, to implement the ResMLP architecture, you can implement the follow model function:\n\n```python\nimport haiku as hk\nimport jax\nfrom einops import rearrange\nfrom x_mlps import XMLP, Affine, resmlp_block_factory\n\ndef create_model(patch_size: int, dim: int, depth: int, num_classes: int = 10):\n    def model_fn(x: jnp.ndarray, is_training: bool) -> jnp.ndarray:\n        # Reformat input image into a sequence of patches\n        x = rearrange(x, "(h p1) (w p2) c -> (h w) (p1 p2 c)", p1=patch_size, p2=patch_size)\n        return XMLP(\n            num_patches=x.shape[-2],\n            dim=dim,\n            depth=depth,\n            block=resmlp_block_factory,\n            # Normalization following the stack of ResMLP blocks\n            normalization=lambda num_patches, dim, depth, **kwargs: Affine(dim, **kwargs),\n            num_classes=num_classes,\n        )(x, is_training=is_training)\n\n    # NOTE: Operating directly on batched data is supported as well.\n    return hk.vmap(model_fn, in_axes=(0, None))\n\nmodel = create_model(patch_size=4, dim=384, depth=12)\nmodel_fn = hk.transform(model)\n\nrng = jax.random.PRNGKey(0)\nparams = model_fn.init(rng, jnp.ones((1, 32, 32, 3)), False)\n```\n\nIt\'s important to note the `XMLP` module _does not_ reformat input data for you (e.g., to a sequence of patches).\nAs such, you must reformat data manually before feeding it to an `XMLP` module.\nThe [einops](https://github.com/arogozhnikov/einops) library, which is installed by X-MLPs, provides functions that can help here (e.g., `rearrange`).\n\n**Note**: Like the core Haiku modules, all modules implemented in X-MLPs support batched data and being vectorized via `vmap`.\n\n## X-MLPs Architecture Details\n\nX-MLPs uses a layered approach to construct arbitrary MLP networks. There are three core modules used to create a network\'s structure:\n\n1. `XSublayer` - bottom level module which wraps arbitrary feedforward functionality.\n2. `XBlock` - mid level module consisting of one or more `XSublayer` modules.\n3. `XMLP` - top level module which represents a generic MLP network, and is composed of a stack of repeated `XBlock` modules.\n\nTo support user-defined modules, each of the above modules support passing arbitrary keyword arguments to child modules.\nThis is accomplished by prepending arguments with one or more predefined prefixes (including user defined prefixes).\nBuilt-in prefixes include:\n\n1. "block\\_" - arguments fed directly to the `XBlock` module.\n2. "sublayers\\_" - arguments fed to all `XSublayer`s in each `XBlock`.\n3. "sublayers{i}\\_" - arguments fed to the i-th `XSublayer` in each `XBlock` (where 1 <= i <= # of sublayers).\n4. "ff\\_" - arguments fed to the feedforward module in a `XSublayer`.\n\nPrefixes must be combined in order when passing them to the `XMLP` module (e.g., "block_sublayer1_ff\\_<argument name>").\n\n### XSublayer\n\nThe `XSublayer` module is a flexible sublayer wrapper module providing skip connections and pre/post-normalization to an arbitrary child module (specifically, arbitrary feedforward modules e.g., `XChannelFeedForward`).\nChild module instances are not passed directly, rather a factory function which creates the child module is instead.\nThis ensures that individual sublayers can be configured automatically based on depth.\n\n### XBlock\n\nThe `XBlock` module is a generic MLP block. It is composed of one or more `XSublayer` modules, passed as factory functions.\n\n### XMLP\n\nAt the top level is the `XMLP` module, which represents a generic MLP network.\nN `XBlock` modules are stacked together to form a network, created via a common factory function.\n\n## Built-in MLP Architectures\n\nThe following architectures have been implemented in the form of `XBlock`s and have corresponding factory functions.\n\n- ResMLP - `resmlp_block_factory`\n- MLP-Mixer - `mlpmixer_block_factory`\n- gMLP - `gmlp_block_factory`\n- S²-MLP - `s2mlp_block_factory`\n- FNet - `fft_block_factory`\n\nImportantly, the components that make up these blocks are part of the public API, and thus can be easily mixed and\nmatched. Additionally, several component variations have been made based on combining ideas from current research.\nThis includes:\n\n- `fftlinear_block_factory` - an FNet block variation: combines a 1D FFT for patch mixing plus a linear layer.\n- `create_shift1d_op` - A 1D shift operation inspired by S²-MLP, making it appropriate for sequence data.\n- `create_multishift1d_op` - Like `create_shift1d_op`, but supports multiple shifts of varying sizes.\n\nSee their respective docstrings for more information.\n\n## LICENSE\n\nSee [LICENSE](LICENSE).\n\n## Citations\n\n```bibtex\n@article{Touvron2021ResMLPFN,\n  title={ResMLP: Feedforward networks for image classification with data-efficient training},\n  author={Hugo Touvron and Piotr Bojanowski and Mathilde Caron and Matthieu Cord and Alaaeldin El-Nouby and Edouard Grave and Gautier Izacard and Armand Joulin and Gabriel Synnaeve and Jakob Verbeek and Herv\'e J\'egou},\n  journal={ArXiv},\n  year={2021},\n  volume={abs/2105.03404}\n}\n```\n\n```bibtex\n@article{Tolstikhin2021MLPMixerAA,\n  title={MLP-Mixer: An all-MLP Architecture for Vision},\n  author={Ilya O. Tolstikhin and Neil Houlsby and Alexander Kolesnikov and Lucas Beyer and Xiaohua Zhai and Thomas Unterthiner and Jessica Yung and Daniel Keysers and Jakob Uszkoreit and Mario Lucic and Alexey Dosovitskiy},\n  journal={ArXiv},\n  year={2021},\n  volume={abs/2105.01601}\n}\n```\n\n```bibtex\n@article{Liu2021PayAT,\n  title={Pay Attention to MLPs},\n  author={Hanxiao Liu and Zihang Dai and David R. So and Quoc V. Le},\n  journal={ArXiv},\n  year={2021},\n  volume={abs/2105.08050}\n}\n```\n\n```bibtex\n@article{Yu2021S2MLPSM,\n  title={S2-MLP: Spatial-Shift MLP Architecture for Vision},\n  author={Tan Yu and Xu Li and Yunfeng Cai and Mingming Sun and Ping Li},\n  journal={ArXiv},\n  year={2021},\n  volume={abs/2106.07477}\n}\n```\n\n```bibtex\n@article{Touvron2021GoingDW,\n  title={Going deeper with Image Transformers},\n  author={Hugo Touvron and Matthieu Cord and Alexandre Sablayrolles and Gabriel Synnaeve and Herv\'e J\'egou},\n  journal={ArXiv},\n  year={2021},\n  volume={abs/2103.17239}\n}\n```\n\n```bibtex\n@inproceedings{Huang2016DeepNW,\n  title={Deep Networks with Stochastic Depth},\n  author={Gao Huang and Yu Sun and Zhuang Liu and Daniel Sedra and Kilian Q. Weinberger},\n  booktitle={ECCV},\n  year={2016}\n}\n```\n\n```bibtex\n@article{LeeThorp2021FNetMT,\n  title={FNet: Mixing Tokens with Fourier Transforms},\n  author={James P Lee-Thorp and Joshua Ainslie and Ilya Eckstein and Santiago Onta{\\~n}{\\\'o}n},\n  journal={ArXiv},\n  year={2021},\n  volume={abs/2105.03824}\n}\n```\n',
    'author': 'Miller Wilt',
    'author_email': 'miller@pyriteai.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/PyriteAI/x-mlps',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
